#!/usr/bin/env python
# vim:fileencoding=utf-8
# License: BSD Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil
import sys
import tempfile
import unittest

base = os.path.dirname(os.path.abspath(__file__))
iswindows = hasattr(sys, 'getwindowsversion')


class TestCase(unittest.TestCase):

    ae = unittest.TestCase.assertEqual
    longMessage = True
    tb_locals = True
    maxDiff = None


class TempDir(object):

    def __enter__(self):
        self.tdir = tempfile.mkdtemp()
        if isinstance(self.tdir, bytes):
            self.tdir = self.tdir.decode(sys.getfilesystemencoding())
        return self.tdir

    def __exit__(self, *a):
        shutil.rmtree(self.tdir)
        del self.tdir



#!/usr/bin/env python2
# vim:fileencoding=utf-8
# License: BSD Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>

from __future__ import absolute_import, division, print_function, unicode_literals

import hashlib
import os
import unicodedata
import unittest
from binascii import crc32

from unrardll import (
    BadPassword, PasswordRequired, comment, extract, extract_member, headers, names,
    open_archive, unrar
)

from . import TempDir, TestCase, base

multipart_rar = os.path.join(base, 'example_split_archive.part1.rar')
password_rar = os.path.join(base, 'example_password_protected.rar')
simple_rar = os.path.join(base, 'simple.rar')
sr_data = {
    '1': b'',
    '1/sub-one': b'sub-one\n',
    '2': b'',
    '2/sub-two.txt': b'sub-two\n',
    'Füße.txt': b'unicode\n',
    'max-compressed': b'max\n',
    'one.txt': b'one\n',
    'symlink': b'sub-two',
    'uncompressed': b'uncompressed\n',
    '诶比屁.txt': b'chinese unicode\n'}


def normalize(x):
    return unicodedata.normalize('NFC', x)


def get_memory():
    'Return memory usage in bytes'
    # See https://pythonhosted.org/psutil/#psutil.Process.memory_info
    try:
        import psutil
    except ImportError:
        raise unittest.SkipTest('psutil is not available')
    return psutil.Process(os.getpid()).memory_info().rss


def memory(since=0.0):
    'Return memory used in MB. The value of since is subtracted from the used memory'
    ans = get_memory()
    ans /= float(1024**2)
    return ans - since


class BasicTests(TestCase):

    def test_names(self):
        all_names = [
            '1/sub-one', 'one.txt', '诶比屁.txt', 'Füße.txt', '2/sub-two.txt',
            'symlink', '1', '2', 'uncompressed', 'max-compressed']
        self.ae(all_names, list(names(simple_rar)))
        all_names.remove('symlink'), all_names.remove('1'), all_names.remove('2')
        self.ae(all_names, list(names(simple_rar, only_useful=True)))

    def test_comment(self):
        self.ae(comment(simple_rar), 'some comment\n')
        self.ae(comment(password_rar), '')

    def test_share_open(self):
        with open(simple_rar, 'rb') as f:
            self.ae(comment(simple_rar), 'some comment\n')
            f.close()

    def test_extract(self):
        for v in (True, False):
            with TempDir() as tdir:
                extract(simple_rar, tdir, verify_data=v)
                h = {
                    normalize(os.path.abspath(os.path.join(tdir, h['filename']))): h
                    for h in headers(simple_rar)}
                data = {}
                for dirpath, dirnames, filenames in os.walk(tdir):
                    for f in filenames:
                        path = normalize(os.path.join(dirpath, f))
                        data[os.path.relpath(path, tdir).replace(os.sep, '/')] = d = open(path, 'rb').read()
                        if f == 'one.txt':
                            self.ae(os.path.getmtime(path), 1098472879)
                        self.ae(h[path]['unpack_size'], len(d))
                        self.ae(h[path]['file_crc'] & 0xffffffff, crc32(d) & 0xffffffff)
            q = {k: v for k, v in sr_data.items() if v}
            del q['symlink']
            self.ae(data, q)

    def test_password(self):
        with TempDir() as tdir:
            self.assertRaises(PasswordRequired, extract, password_rar, tdir)
            self.assertRaises(BadPassword, extract, password_rar, tdir, password='sfasgsfdg')
            extract(password_rar, tdir, password='example')

    def test_invalid_callback(self):
        class Callback(object):
            pass
        with open_archive(simple_rar, None, mode=unrar.RAR_OM_EXTRACT) as f:
            unrar.read_next_header(f)
            self.assertRaisesRegexp(unrar.UNRARError, "An exception occurred in the password callback handler", unrar.process_file, f)
        c = Callback()
        c._process_data = lambda x: None
        with open_archive(simple_rar, c, mode=unrar.RAR_OM_EXTRACT) as f:
            unrar.read_next_header(f)
            self.assertRaisesRegexp(unrar.UNRARError,  "Processing canceled by the callback", unrar.process_file, f)

    def test_multipart(self):
        self.ae(list(names(multipart_rar)), ['Fifteen_Feet_of_Time.pdf'])
        for v in (True, False):
            with TempDir() as tdir:
                extract(multipart_rar, tdir, verify_data=v)
                h = next(headers(multipart_rar))
                raw = open(os.path.join(tdir, h['filename']), 'rb').read()
                self.ae(len(raw), h['unpack_size'])
                self.ae(hashlib.sha1(raw).hexdigest(), 'a9fc6a11d000044f17fcdf65816348ce0be3b145')

    def test_extract_member(self):
        self.ae(extract_member(simple_rar, lambda h: h['filename'] == 'one.txt', verify_data=True), ('one.txt', sr_data['one.txt']))
        self.ae(extract_member(simple_rar, lambda h: False), (None, None))

    def test_open_failure(self):
        self.assertRaises(OSError, extract, 'sdfgsfgsggsdfg.rar', '.')

    def test_memory_leaks(self):
        import gc

        def collect():
            for i in range(6):
                gc.collect()
            gc.collect()

        with TempDir() as tdir:

            def get_mem_use(num):
                collect()
                start = memory()
                for i in range(num):
                    extract(simple_rar, tdir)
                collect()
                return max(0, memory(start))

            get_mem_use(5)  # ensure no memory used by get_mem_use itself is counted
            a, b = get_mem_use(10), get_mem_use(100)
        self.assertTrue(a == 0 or b/a < 3, '10 times usage: {} 100 times usage: {}'.format(a, b))



#!/usr/bin/env python
# vim:fileencoding=utf-8
# License: BSD Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>

from __future__ import absolute_import, division, print_function, unicode_literals

import errno
import os
import sys
from binascii import crc32
from collections import namedtuple, defaultdict
from contextlib import contextmanager

from . import unrar

V = namedtuple('Version', 'major minor patch')

version = V(0, 1, 4)
RARDLL_VERSION = unrar.RARDllVersion
iswindows = hasattr(sys, 'getwindowsversion')
isosx = 'darwin' in sys.platform.lower()

# local_open() opens a file that wont be inherited by child processes  {{{
if sys.version_info.major < 3:
    if iswindows:
        def local_open(name, mode='r', bufsize=-1):
            mode += 'N'
            return open(name, mode, bufsize)
    elif isosx:
        import fcntl
        FIOCLEX = 0x20006601

        def local_open(name, mode='r', bufsize=-1):
            ans = open(name, mode, bufsize)
            try:
                fcntl.ioctl(ans.fileno(), FIOCLEX)
            except EnvironmentError:
                fcntl.fcntl(ans, fcntl.F_SETFD, fcntl.fcntl(ans, fcntl.F_GETFD) | fcntl.FD_CLOEXEC)
            return ans
    else:
        import fcntl
        try:
            cloexec_flag = fcntl.FD_CLOEXEC
        except AttributeError:
            cloexec_flag = 1
        supports_mode_e = False

        def local_open(name, mode='r', bufsize=-1):
            global supports_mode_e
            mode += 'e'
            ans = open(name, mode, bufsize)
            if supports_mode_e:
                return ans
            old = fcntl.fcntl(ans, fcntl.F_GETFD)
            if not (old & cloexec_flag):
                fcntl.fcntl(ans, fcntl.F_SETFD, old | cloexec_flag)
            else:
                supports_mode_e = True
            return ans
else:
    local_open = open
# }}}


def is_useful(h):
    return not (h['is_dir'] or h['redir_type'])


class Callback(object):

    def __init__(self, pw=None):
        self.pw = type('')(pw) if pw is not None else None
        self.password_requested = False

    def _get_password(self):
        self.password_requested = True
        return self.pw

    def _process_data(self, data):
        pass

    def reset(self):
        self.password_requested = False


def safe_path(base, relpath):
    base = os.path.abspath(base)
    path = os.path.abspath(os.path.join(base, relpath))
    if (
        os.path.normcase(path) == os.path.normcase(base) or
        not os.path.normcase(path).startswith(os.path.normcase(base))
    ):
        return None
    return path


def is_safe_symlink(base, x):
    base = os.path.normcase(base)
    tgt = os.path.abspath(os.path.join(base, x))
    ntgt = os.path.normcase(tgt)
    extra = ntgt[len(base):]
    return ntgt.startswith(base) and (not extra or extra[0] in (os.sep, '/'))


def ensure_dir(path):
    try:
        os.makedirs(path)
    except EnvironmentError as err:
        if err.errno != errno.EEXIST:
            raise


class PasswordError(ValueError):
    pass


class PasswordRequired(PasswordError):

    def __init__(self, archive_path):
        ValueError.__init__(self, 'A password is required for: %r' % archive_path)


class BadPassword(PasswordError):

    def __init__(self, archive_path):
        ValueError.__init__(self, 'The specified password is incorrect for: %r' % archive_path)


def do_func(func, archive_path, f, c, *args):
    try:
        return func(f, *args)
    except unrar.UNRARError as e:
        m = e.args[0]
        if m == 'ERAR_MISSING_PASSWORD':
            raise PasswordRequired(archive_path)
        if m == 'ERAR_BAD_DATA' and c.password_requested:
            raise (BadPassword if c.pw else PasswordRequired)(archive_path)
        raise


@contextmanager
def open_archive(archive_path, callback, mode=unrar.RAR_OM_LIST, get_comment=False):
    try:
        f = unrar.open_archive(archive_path, callback, mode, get_comment)
        if get_comment:
            f, c = f
    except unrar.UNRARError as e:
        m = e.args[0]
        raise OSError((errno.ENOENT, 'Failed to open archive at: %r with underlying unrar error code: %s' % (
            archive_path, m), archive_path))
    yield (f, c) if get_comment else f
    unrar.close_archive(f)
    del f


def headers(archive_path, password=None, mode=unrar.RAR_OM_LIST):
    ''' Yield the headers for all files in the archive '''
    c = Callback(pw=password)
    archive_path = type('')(archive_path)
    with open_archive(archive_path, c, mode) as f:
        while True:
            h = do_func(unrar.read_next_header, archive_path, f, c)
            if h is None:
                break
            yield h
            do_func(unrar.process_file, archive_path, f, c)
            c.reset()


def names(archive_path, only_useful=False, password=None):
    ''' Yield the archive file names for all files in the archive '''
    for h in headers(archive_path, password=password):
        if not only_useful or is_useful(h):
            yield h['filename'].replace(os.sep, '/')


def comment(archive_path):
    ''' Read the archive comment, if any. Return empty string if no comment. '''
    c = Callback()
    archive_path = type('')(archive_path)
    with open_archive(archive_path, c, get_comment=True) as x:
        # dll.cpp in the unrar source code must be patched, replacing WideToChar
        # with WideToUtf otherwise the comment could be in any old system
        # dependent encoding.
        return x[1].decode('utf-8')


class ExtractCallback(Callback):

    def __init__(self, pw=None, verify_data=False):
        self.verify_data = verify_data
        Callback.__init__(self, pw=pw)
        self.crc = 0

    def _process_data(self, data):
        self.write(data)
        self.written += len(data)
        if self.verify_data:
            self.crc = crc32(data, self.crc) & 0xffffffff
        return True

    def reset(self, write=None, crc=0):
        Callback.reset(self)
        self.written = 0
        self.write = write
        self.crc = crc


class FileCorrupt(ValueError):
    pass


def verify(archive_path, crc_map, password=None):
    # Verify CRCs
    crcs = {}
    for h in headers(archive_path, password=password, mode=unrar.RAR_OM_LIST_INCSPLIT):
        crcs[h['filename']] = h['file_crc']
    for k in crc_map:
        got = crc_map[k] & 0xffffffff
        nominal = crcs.get(k, 0) & 0xffffffff
        if nominal != got:
            raise FileCorrupt('The CRC for %r does not match. Expected: %d Got %d' % (
                k, nominal, got))


def _extract(f, archive_path, c, location):
    seen = set()
    crc_map = defaultdict(lambda: 0)
    while True:
        h = do_func(unrar.read_next_header, archive_path, f, c)
        if h is None:
            break
        filename = h['filename']
        if not filename:
            continue
        open_file = None
        dest = safe_path(location, filename)
        c.reset(crc=crc_map[filename])
        extracted = False
        if h['is_dir']:
            try:
                os.makedirs(safe_path(location, filename))
            except Exception:
                pass
                # We ignore create directory errors since we dont
                # care about missing empty dirs
            crc_map.pop(filename)
        elif h['redir_type'] != 0:
            if h['redir_type'] == 1:  # Unix symlink
                syn = h.get('redir_name')
                if syn and not iswindows:
                    # Only RAR 5 archives have a redir_name
                    syn_base = os.path.dirname(dest)
                    if is_safe_symlink(location, os.path.join(syn_base, syn)):
                        ensure_dir(syn_base)
                        os.symlink(syn, dest)
            crc_map.pop(filename)
        else:
            ensure_dir(os.path.dirname(dest))
            open_file = local_open(dest, 'ab' if dest in seen else 'wb')
            c.reset(write=open_file.write, crc=crc_map[filename])
            extracted = True
        try:
            if open_file is not None and not c.verify_data:
                do_func(unrar.process_file, archive_path, f, c, unrar.RAR_TEST, open_file.fileno())
            else:
                do_func(unrar.process_file, archive_path, f, c)
        finally:
            if open_file is not None:
                open_file.close()
        seen.add(dest)
        if extracted:
            crc_map[filename] = c.crc
            c.reset()  # so that file is closed
            os.utime(dest, (h['file_time'], h['file_time']))
    return crc_map


def extract(archive_path, location='.', password=None, verify_data=False):
    ''' Extract all files from the archive to the specified location, which must be an existing directory. '''
    c = ExtractCallback(pw=password, verify_data=verify_data)
    archive_path = type('')(archive_path)
    with open_archive(archive_path, c, unrar.RAR_OM_EXTRACT) as f:
        crc_map = _extract(f, archive_path, c, location)
    del f
    if verify_data:
        verify(archive_path, crc_map, password=password)


def extract_member(archive_path, predicate, password=None, verify_data=False):
    ''' Extract a single file from the archive for which the predicate function returns true. Return (file name, data as bytes). '''
    c = ExtractCallback(pw=password, verify_data=verify_data)
    archive_path = type('')(archive_path)
    with open_archive(archive_path, c, unrar.RAR_OM_EXTRACT) as f:
        while True:
            h = do_func(unrar.read_next_header, archive_path, f, c)
            if h is None:
                return None, None
            if h['is_dir'] or h['redir_type'] or not predicate(h):
                do_func(unrar.process_file, archive_path, f, c, unrar.RAR_SKIP)
            else:
                buf = []
                c.reset(write=buf.append)
                do_func(unrar.process_file, archive_path, f, c)
                break
    del f
    crc_map = {h['filename']: c.crc}
    if verify_data:
        verify(archive_path, crc_map, password=password)
    return h['filename'], b''.join(buf)



#!/usr/bin/env python
# vim:fileencoding=utf-8
# License: BSD Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>

from __future__ import absolute_import, division, print_function, unicode_literals

import glob
import os
import re
import shutil
import subprocess
import sys
import tarfile
import time
from io import BytesIO

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

isosx = 'darwin' in sys.platform.lower()
iswindows = hasattr(sys, 'getwindowsversion')
is64bit = sys.maxsize > (1 << 32)


def download(url):
    i = 5
    while i > 0:
        i -= 1
        try:
            return urlopen(url).read()
        except Exception:
            if i <= 0:
                raise
            print('Download failed, retrying...')
            sys.stdout.flush()
            time.sleep(1)


def download_unrar():
    html = download('https://www.rarlab.com/rar_add.htm').decode('utf-8', 'replace')
    href = re.search(r'<a\s+.*?href="([^"]+)".*?>UnRAR source</a>', html).group(1)
    href = 'https://www.rarlab.com/' + href
    print('Downloading unrar', href)
    sys.stdout.flush()
    return download(href)


def download_and_extract():
    raw = download_unrar()
    with tarfile.open(fileobj=BytesIO(raw), mode='r:*') as tf:
        tf.extractall()


def replace_in_file(path, old, new, missing_ok=False):
    if isinstance(old, type('')):
        old = old.encode('utf-8')
    if isinstance(new, type('')):
        new = new.encode('utf-8')
    with open(path, 'r+b') as f:
        raw = f.read()
        if isinstance(old, bytes):
            nraw = raw.replace(old, new)
        else:
            nraw = old.sub(new, raw)
        if raw == nraw and not missing_ok:
            raise ValueError('Failed (pattern not found) to patch: ' + path)
        f.seek(0), f.truncate()
        f.write(nraw)


def build_unix():
    if isosx:
        with open('makefile', 'r+b') as m:
            raw = m.read().decode('utf-8')
            raw = raw.replace('libunrar.so', 'libunrar.dylib')
            m.seek(0), m.truncate()
            m.write(raw.encode('utf-8'))
    flags = '-fPIC ' + os.environ.get('CXXFLAGS', '')
    subprocess.check_call(['make', '-j4', 'lib', 'CXXFLAGS="%s"' % flags.strip()])
    lib = 'libunrar.' + ('dylib' if isosx else 'so')
    shutil.copy2(lib, '../../lib')


def build_windows():
    PL = 'x64' if is64bit else 'Win32'
    subprocess.check_call([
        'msbuild.exe', 'UnRARDll.vcxproj', '/t:Build', '/p:Platform=' + PL, '/p:Configuration=Release'])
    lib = glob.glob('./build/*/Release/UnRAR.lib')[0]
    dll = glob.glob('./build/*/Release/unrar.dll')[0]
    shutil.copy2(lib, '../../lib')
    shutil.copy2(dll, '../../..')


def build_unrar():
    os.makedirs('sw/build'), os.makedirs('sw/include/unrar'), os.mkdir('sw/lib')
    os.chdir('sw/build')
    download_and_extract()
    os.chdir('unrar')
    replace_in_file('dll.cpp', 'WideToChar', 'WideToUtf')
    (build_windows if iswindows else build_unix)()
    for f in glob.glob('*.hpp'):
        shutil.copy2(f, '../../include/unrar')


if __name__ == '__main__':
    build_unrar()



#!/usr/bin/env python2
# vim:fileencoding=utf-8
# License: BSD Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>

from __future__ import absolute_import, division, print_function, unicode_literals

import importlib
import os
import re
import sys
from distutils.command.build import build as Build

from setuptools import Extension, setup

self_path = os.path.abspath(__file__)
base = os.path.dirname(self_path)
iswindows = hasattr(sys, 'getwindowsversion')
raw = open(os.path.join(base, 'src/unrardll/__init__.py'),
           'rb').read().decode('utf-8')
version = map(
    int, re.search(r'^version = V\((\d+), (\d+), (\d+)', raw, flags=re.M).groups())


def include_dirs():
    ans = []
    if 'UNRAR_INCLUDE' in os.environ:
        ans.extend(os.environ['UNRAR_INCLUDE'].split(os.pathsep))
    return ans


def libraries():
    return ['unrar']


def library_dirs():
    ans = []
    if 'UNRAR_LIBDIRS' in os.environ:
        ans.extend(os.environ['UNRAR_LIBDIRS'].split(os.pathsep))
    return ans


def macros():
    ans = [
        ('SILENT', 1),
        ('RARDLL', 1),
        ('UNRAR', 1), ]
    if not iswindows:
        ans.append(('_UNIX', 1))
    return ans


def find_tests():
    import unittest
    suites = []
    for f in os.listdir(os.path.join(base, 'test')):
        n, ext = os.path.splitext(f)
        if ext == '.py' and n not in ('__init__',):
            m = importlib.import_module('test.' + n)
            suite = unittest.defaultTestLoader.loadTestsFromModule(m)
            suites.append(suite)
    return unittest.TestSuite(suites)


class Test(Build):

    description = "run unit tests after in-place build"

    def run(self):
        import unittest
        Build.run(self)
        if self.dry_run:
            self.announce('skipping "test" (dry run)')
            return
        sys.path.insert(0, self.build_lib)
        tests = find_tests()
        r = unittest.TextTestRunner
        result = r(verbosity=2).run(tests)

        if not result.wasSuccessful():
            raise SystemExit(1)


CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
License :: OSI Approved :: BSD License
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python
Topic :: Software Development :: Libraries :: Python Modules
Topic :: System :: Archiving :: Compression
"""

setup(
    name=str('unrardll'),
    version='{}.{}.{}'.format(*version),
    author='Kovid Goyal',
    author_email='redacted@acme.com',
    description='Wrap the Unrar DLL to enable unraring of files in python',
    license='BSD',
    url='https://github.com/kovidgoyal/unrardll',
    classifiers=[c for c in CLASSIFIERS.split("\n") if c],
    platforms=['any'],
    packages=[str('unrardll')],
    package_dir={'': str('src')},
    cmdclass={'test': Test},
    ext_modules=[
        Extension(
            str('unrardll.unrar'),
            include_dirs=include_dirs(),
            libraries=libraries(),
            library_dirs=library_dirs(),
            define_macros=macros(),
            sources=[str('src/unrardll/wrapper.cpp')]
        )
    ]
)
